from typing import Iterable

from prometheus_client import Metric
from querent.common.errors.metric_errors import UnknownMetricError, UpdateMetricError
from querent.metric.base_metric_adapter import MetricAdapter


class MetricRegistry:
    def __init__(self):
        self.metric_adapters: dict[str, MetricAdapter] = {}

    def register_metric_adapter(self, metric_name: str, metric_adapter: MetricAdapter):
        """
        Register a metric adapter.
        :param metric_adapter: Metric adapter instance.
        """
        self.metric_adapters[metric_name.lower()] = metric_adapter

    def list_metric_adapters(self):
        """
        List all registered metric adapters.
        :return: List of metric adapter names.
        """
        return list(self.metric_adapters.keys())

    def update_metric(self, metric_name: str, value: int):
        """
        Update the value of a metric.
        :param metric_name: Name of the metric.
        :param value: New value for the metric.
        """
        try:
            self.metric_adapters[metric_name.lower()].update_metric(value)
        except Exception as e:
            raise UpdateMetricError(f"Metric {metric_name} is not registered.") from e

    def get_metrics(self, metric_name: str) -> Iterable[Metric]:
        """
        Get the value of a metric.
        :param metric_name: Name of the metric.
        :return: Value of the metric.
        """
        try:
            return self.metric_adapters[metric_name.lower()].get_metric()
        except KeyError:
            raise UnknownMetricError(f"Metric {metric_name} is not registered.")



from abc import ABC, abstractmethod
from typing import Any

from querent.common.types.querent_event import EventState, EventType
from querent.common.types.ingested_tokens import IngestedTokens
from querent.logging.logger import setup_logger

logger = setup_logger(__name__, "ChannelCommandInterface")


class ChannelCommandInterface(ABC):
    @abstractmethod
    def receive_in_python(self) -> EventState:
        """Receive a message in Python from Rust."""
        raise NotImplementedError

    @abstractmethod
    def send_in_rust(self, message_type: EventType, message_data: Any) -> None:
        """Send a message in Rust from Python."""
        raise NotImplementedError
    
    @abstractmethod
    def receive_tokens_in_python(self) -> IngestedTokens:
        """Receive tokens in Python from Rust."""
        raise NotImplementedError
    
    @abstractmethod
    def send_tokens_in_rust(self, tokens: IngestedTokens) -> None:
        """Send tokens in Rust from Python."""
        raise NotImplementedError



from typing import Iterable
from prometheus_client import Metric


class MetricAdapter:
    def __init__(self):
        raise NotImplementedError

    def update_metric(self, metric_name, value):
        raise NotImplementedError

    def get_metric(self, metric_name) -> Iterable[Metric]:
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import Any

from querent.common.types.querent_event import EventState, EventType
from querent.logging.logger import setup_logger

logger = setup_logger(__name__, "EventCallbackInterface")


class EventCallbackInterface(ABC):
    @abstractmethod
    def handle_event(self, event_type: EventType, event_data: EventState):
        """
        Handle an event.
        Args:
            event_type (EventType): The type of event to subscribe to (e.g., "token_processed").
            event_data (EventState): Data associated with the event.

        """

        raise NotImplementedError



import json
from typing import AsyncGenerator

from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.common import common_errors
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.uri import Uri
from querent.config.collector.collector_config import (
    CollectorBackend,
    JiraCollectorConfig,
)
from jira import JIRA

from querent.logging.logger import setup_logger


class JiraCollector(Collector):
    def __init__(self, config: JiraCollectorConfig):
        self.config = config
        self.jira = None  # Initialize to None
        self.logger = setup_logger(__name__, "JiraCollector")

    def convert_to_boolean(self, val: str):
        if type(val) == str:
            return val.lower() == "true"
        return val

    def create_jira_client(self):
        self.config.jira_verify = self.convert_to_boolean(self.config.jira_verify)
        options = {
            "server": self.config.jira_server,
            "verify": self.config.jira_verify,
        }
        basic_auth = None
        token_auth = None
        if self.config.jira_api_token:
            basic_auth = (self.config.jira_username, self.config.jira_api_token)
        elif self.config.jira_password:
            token_auth = (self.config.jira_username, self.config.jira_password)
        elif self.config.jira_keyfile and self.config.jira_certfile:
            options["client_cert"] = (
                self.config.jira_certfile,
                self.config.jira_keyfile,
            )

        return JIRA(
            self.config.jira_server,
            basic_auth=basic_auth,
            token_auth=token_auth,
            options=options,
        )

    async def connect(self):
        try:
            self.jira = self.create_jira_client()
        except Exception as e:
            self.logger.error(f"Error connecting to Jira: {e}")
            raise common_errors.ConnectionError(
                f"Failed to connect to Jira: {str(e)}"
            ) from e

    async def disconnect(self):
        self.jira = None

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        try:
            if not self.jira:
                raise common_errors.ConnectionError(
                    "Jira client not initialized. Call connect() before polling."
                )

            issues = self.jira.search_issues(
                self.config.jira_query,
                startAt=self.config.jira_start_at,
                maxResults=self.config.jira_max_results,
                fields=self.config.jira_fields,
                expand=self.config.jira_expand,
                json_result=False,
            )

            for issue in issues:
                json_issue = json.dumps(issue.raw).encode("utf-8")
                yield CollectedBytes(
                    data=json_issue, file=f"jira_issue_{issue.key}.json.jira"
                )
                yield CollectedBytes(
                    data=None, file=f"jira_issue_{issue.key}.json.jira", eof=True
                )

        except common_errors.ConnectionError as e:
            self.logger.error(f"Error polling Jira issues: {e}")
            raise  # Re-raise ConnectionError without adding additional information
        except Exception as e:
            raise common_errors.ConnectionError(
                f"Failed to poll Jira issues: {str(e)}"
            ) from e
        finally:
            await self.disconnect()


class JiraCollectorFactory(CollectorFactory):
    def backend(self) -> CollectorBackend:
        return CollectorBackend.Jira

    def resolve(self, uri: Uri, config: JiraCollectorConfig) -> Collector:
        return JiraCollector(config)



from abc import abstractmethod
from typing import AsyncGenerator, List
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.types.ingested_tokens import IngestedTokens
from querent.processors.async_processor import AsyncProcessor


class BaseIngestor:
    def __init__(self, processors: List[AsyncProcessor]):
        self.processors = processors

    @abstractmethod
    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        # Your common ingestion logic here
        raise NotImplementedError

    @abstractmethod
    async def process_data(self, text):
        # Your common data processing logic here
        raise NotImplementedError



import asyncio
from aiohttp import ClientSession, TCPConnector
from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.common.types.collected_bytes import CollectedBytes
from querent.config.collector.collector_config import CollectorBackend, WebScraperConfig
from querent.common.uri import Uri
from querent.tools.web_page_extractor import WebpageExtractor
from urllib.parse import urlparse, urljoin


class WebScraperCollector(Collector):
    def __init__(self, config: WebScraperConfig):
        self.website_url = config.website_url
        self.semaphore = asyncio.Semaphore(
            5
        )  # Adjust the limit as needed (e.g., 5 requests at a time)
        self.poll_lock = asyncio.Lock()  # Lock for the poll method

    async def connect(self):
        pass  # Any setup logic before scraping

    async def disconnect(self):
        pass  # Any cleanup logic after scraping

    async def poll(self):
        async with self.poll_lock:
            urls_to_scrape = [self.website_url]
            while urls_to_scrape:
                url = urls_to_scrape.pop()
                content = await self.scrape_website(url)
                yield CollectedBytes(file=None, data=content.data, error=None)
                # Find and add links from this page to the list of URLs to scrape
                new_urls = self.extract_links(url)
                urls_to_scrape.extend(new_urls)

    async def scrape_website(self, website_url: str):
        async with self.semaphore:
            async with ClientSession(connector=TCPConnector(ssl=False)) as session:
                async with session.get(website_url) as response:
                    content = await response.text()
                    max_length = len(content)
                    return CollectedBytes(
                        data=content[:max_length], file=None, error=None
                    )

    def extract_links(self, base_url: str):
        # Use a proper HTML parser to extract links
        extractor = WebpageExtractor()
        links = extractor.extract_links(base_url)
        # Join relative links with the base URL
        return [urljoin(base_url, link) for link in links]


class WebScraperFactory(CollectorFactory):
    def __init__(self):
        pass

    def backend(self) -> CollectorBackend:
        return CollectorBackend.WebScraper

    def resolve(self, uri: Uri, config: WebScraperConfig) -> Collector:
        return WebScraperCollector(config)



from typing import AsyncGenerator, List
import json
from querent.common.types.collected_bytes import CollectedBytes
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.ingestors.base_ingestor import BaseIngestor
from querent.ingestors.ingestor_factory import IngestorFactory
from querent.processors.async_processor import AsyncProcessor
from querent.common import common_errors
from querent.common.types.ingested_tokens import IngestedTokens
from querent.logging.logger import setup_logger


class JsonIngestorFactory(IngestorFactory):
    SUPPORTED_EXTENSIONS = {"json", "jira"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return JsonIngestor(processors)


class JsonIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.JSON)
        self.processors = processors
        self.logger = setup_logger(__name__, "JsonIngestor")

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        try:
            current_file = None
            collected_bytes = b""
            try:
                async for chunk_bytes in poll_function:
                    if chunk_bytes.is_error() or chunk_bytes.is_eof():
                        continue

                    if chunk_bytes.file != current_file:
                        if current_file:
                            async for json_objects in self.extract_and_process_json(
                                CollectedBytes(file=current_file, data=collected_bytes)
                            ):
                                yield IngestedTokens(
                                    file=current_file, data=[json_objects], error=None
                                )
                        if current_file:
                            yield IngestedTokens(
                                file=current_file,
                                data=None,
                                error=None,
                            )
                        collected_bytes = b""
                        current_file = chunk_bytes.file

                    collected_bytes += chunk_bytes.data

            except json.JSONDecodeError:
                yield IngestedTokens(
                    file=current_file, data=None, error="JSON Decode Error"
                )
            finally:
                async for json_objects in self.extract_and_process_json(
                    CollectedBytes(file=current_file, data=collected_bytes)
                ):
                    yield IngestedTokens(
                        file=current_file, data=[json_objects], error=None
                    )
                yield IngestedTokens(file=current_file, data=None, error=None)

        except Exception as e:
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")

    async def extract_and_process_json(self, collected_bytes: CollectedBytes) -> str:
        try:
            json_data = collected_bytes.data.decode("utf-8")
            processed_text = await self.process_data(json_data)
            yield processed_text
        except json.JSONDecodeError as exc:
            raise common_errors.JsonDecodeError(
                f"Getting UnicodeDecodeError on this file {collected_bytes.file}"
            ) from exc

        except UnicodeDecodeError as exc:
            raise common_errors.UnicodeDecodeError(
                f"Getting UnicodeDecodeError on this file {collected_bytes.file}"
            ) from exc
        except LookupError as exc:
            raise common_errors.LookupError(
                f"Getting LookupError on this file {collected_bytes.file}"
            ) from exc
        except TypeError as exc:
            raise common_errors.TypeError(
                f"Getting TypeError on this file {collected_bytes.file}"
            ) from exc

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return text
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")



import logging
from logging import LogRecord
from querent.common.errors.metric_errors import UnknownMetricError

from querent.metric.metric_registry import MetricRegistry


class LoggingMetricHandler(logging.Handler):
    def __init__(self, metric_registry: MetricRegistry, level=logging.INFO):
        super().__init__(level)
        self.metric_registry = metric_registry

    def emit(self, record: LogRecord) -> None:
        # Split the log message to extract metric-related information.
        parts = record.msg.split(",")
        metric_name: str = None
        metric_value: int = None

        for part in parts:
            # Assuming the log message contains "M/metric: <metric_name>, V/value: <metric_value>"
            if part.lower().strip().startswith("metric:"):
                metric_name = part.split(":")[1].strip().lower()
            elif part.lower().strip().startswith("value:"):
                metric_value_str = part.split(":")[1].strip().lower()
                if metric_value_str.isdigit():
                    metric_value = int(metric_value_str)

        # Check if metric information was found in the log message.
        if metric_name is not None and metric_value is not None:
            # Check if the metric is registered.
            if metric_name not in self.metric_registry.list_metric_adapters():
                self.logger.error(f"Metric {metric_name} is not registered.")
            # Update the metric value.
            self.metric_registry.update_metric(metric_name, metric_value)
        else:
            # If it's not a metric-related log record, you can handle it as a regular log message.
            super().emit(record)



from typing import List, AsyncGenerator
import io
import speech_recognition as sr
from pydub import AudioSegment

from querent.ingestors.ingestor_factory import IngestorFactory
from querent.processors.async_processor import AsyncProcessor
from querent.ingestors.base_ingestor import BaseIngestor
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.common_errors import (
    UnknownValueError,
    RequestError,
    IndexErrorException,
    UnknownError,
)
from querent.common.types.ingested_tokens import IngestedTokens
from querent.logging.logger import setup_logger


class AudioIngestorFactory(IngestorFactory):
    """
    Factory class for creating AudioIngestor instances.

    This class defines methods to check if a given file extension is supported
    and to create an AudioIngestor object if the extension is supported.
    """

    SUPPORTED_EXTENSIONS = {"mp3"}

    async def supports(self, file_extension: str) -> bool:
        """
        Check if the specified file extension is supported for audio ingestion.

        Args:
            file_extension (str): The file extension to check.

        Returns:
            bool: True if the extension is supported, False otherwise.
        """
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        """
        Create an AudioIngestor instance if the specified extension is supported.

        Args:
            file_extension (str): The file extension to create an ingestor for.
            processors (List[AsyncProcessor]): List of processors to apply to the ingested data.

        Returns:
            BaseIngestor: An AudioIngestor instance or None if the extension is not supported.
        """
        if not await self.supports(file_extension):
            return None
        return AudioIngestor(processors)


class AudioIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.MP3)
        self.processors = processors
        self.logger = setup_logger(__name__, "AudioIngestor")

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        current_file = None
        collected_bytes = b""
        try:
            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    # TODO handle error
                    continue
                if current_file is None:
                    current_file = chunk_bytes.file
                elif current_file != chunk_bytes.file:
                    # we have a new file, process the old one
                    async for text in self.extract_and_process_audio(
                        CollectedBytes(file=current_file, data=collected_bytes)
                    ):
                        yield IngestedTokens(file=current_file, data=[text], error=None)
                    yield IngestedTokens(
                        file=current_file,
                        data=None,
                        error=None,
                    )
                    collected_bytes = b""
                    current_file = chunk_bytes.file
                collected_bytes += chunk_bytes.data
        except Exception as e:
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")
        finally:
            # process the last file
            async for text in self.extract_and_process_audio(
                CollectedBytes(file=current_file, data=collected_bytes)
            ):
                yield IngestedTokens(file=current_file, data=[text], error=None)

            yield IngestedTokens(file=current_file, data=None, error=None)

    async def extract_and_process_audio(
        self, collected_bytes: CollectedBytes
    ) -> AsyncGenerator[IngestedTokens, None]:
        # Recognize the text using the recognizer
        try:
            audio_segment = AudioSegment.from_file(
                io.BytesIO(collected_bytes.data), format=collected_bytes.extension
            )
            temp_wave = io.BytesIO()
            audio_segment.export(temp_wave, format="wav")
            # Initialize the recognizer
            recognizer = sr.Recognizer()

            temp_wave.seek(0)
            with sr.AudioFile(temp_wave) as source:
                recognizer.adjust_for_ambient_noise(source, duration=1)
                audio_data = recognizer.record(source)

            # Recognize the text using the recognizer

            recognized_text = recognizer.recognize_google(audio_data, language="en-US")
            processed_text = await self.process_data(recognized_text)
            yield processed_text
        except sr.UnknownValueError as exc:
            raise UnknownValueError(
                f"The following file gave Unknown Value Error {collected_bytes.file}"
            ) from exc
        except sr.RequestError as exc:
            raise RequestError(
                f"The following file gave Request Error {collected_bytes.file}"
            ) from exc
        except IndexError as exc:
            raise IndexErrorException(
                f"The following file gave Request Error {collected_bytes.file}"
            ) from exc
        except Exception as exc:
            raise UnknownError(
                f"Received unknown error {exc} from the file {collected_bytes.file}"
            ) from exc

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return text
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")



from typing import List, AsyncGenerator
from io import BytesIO
from pptx import Presentation
from pptx.exc import InvalidXmlError
from tika import parser
from zipfile import BadZipFile
from pptx.enum.shapes import MSO_SHAPE_TYPE
import io
from PIL import Image
import pytesseract
import pybase64
import uuid

from querent.ingestors.ingestor_factory import IngestorFactory
from querent.processors.async_processor import AsyncProcessor
from querent.ingestors.base_ingestor import BaseIngestor
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.common.types.collected_bytes import CollectedBytes
from querent.common import common_errors
from querent.common.types.ingested_tokens import IngestedTokens
from querent.common.types.ingested_images import IngestedImages
from querent.logging.logger import setup_logger


class PptIngestorFactory(IngestorFactory):
    SUPPORTED_EXTENSIONS = {"ppt", "pptx"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return PptIngestor(processors)


class PptIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.PPT)
        self.processors = processors
        self.logger = setup_logger(__name__, "PptIngestor")

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        current_file = None
        collected_bytes = b""
        try:
            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    continue

                if current_file is None:
                    current_file = chunk_bytes.file
                elif current_file != chunk_bytes.file:
                    async for ingested_data in self.extract_and_process_ppt(
                        CollectedBytes(file=current_file, data=collected_bytes)
                    ):
                        yield ingested_data
                    yield IngestedTokens(
                        file=current_file,
                        data=None,
                        error=None,
                    )
                    collected_bytes = b""
                    current_file = chunk_bytes.file
                collected_bytes += chunk_bytes.data
        except Exception as e:
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")
        finally:
            async for ingested_data in self.extract_and_process_ppt(
                CollectedBytes(file=current_file, data=collected_bytes)
            ):
                yield ingested_data
            yield IngestedTokens(file=current_file, data=None, error=None)

    async def extract_and_process_ppt(
        self, collected_bytes: CollectedBytes
    ) -> AsyncGenerator[str, None]:
        try:
            if collected_bytes.extension == "pptx":
                ppt_file = BytesIO(collected_bytes.data)
                presentation = Presentation(ppt_file)
                i=1
                for slide in presentation.slides:
                    text = []
                    for shape in slide.shapes:
                        if hasattr(shape, "text"):
                            text.append(shape.text)
                        if shape.shape_type == MSO_SHAPE_TYPE.PICTURE or (shape.shape_type in [MSO_SHAPE_TYPE.PLACEHOLDER, MSO_SHAPE_TYPE.AUTO_SHAPE] and hasattr(shape, "image")):
                            ocr_text = await self.process_image(shape)
                            yield IngestedImages(file=collected_bytes.file, image=pybase64.b64encode(shape.image.blob), image_name=str(uuid.uuid4()), page_num=i, text = text, ocr_text=[ocr_text], coordinates=None)
                    slide_text = "\n".join(text)
                    processed_slide_text = await self.process_data(slide_text)
                    yield IngestedTokens(
                        file=collected_bytes.file, data=processed_slide_text, error=None
                    )
                    i+=1
            elif collected_bytes.extension == "ppt":
                parsed = parser.from_buffer(collected_bytes.data)
                extracted_text = parsed["content"]
                yield IngestedTokens(
                    file=collected_bytes.file, data=extracted_text, error=None
                )
            else:
                raise common_errors.WrongPptFileError(
                    f"Given file is not ppt {collected_bytes.file}"
                )
        except InvalidXmlError as exc:
            raise common_errors.InvalidXmlError(
                f"The following file is not in proper xml format {collected_bytes.file}"
            ) from exc
        except BadZipFile as exc:
            raise common_errors.BadZipFile(
                f"The following file is not a zip file{collected_bytes.file}"
            ) from exc
        except Exception as exc:
            raise common_errors.UnknownError(
                f"Received Unknown error as {exc} from file {collected_bytes.file}"
            ) from exc

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return [text]
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")

    async def process_image(self, shape):
        try:
            # Retrieve image as a BytesIO object
            image_stream = io.BytesIO(shape.image.blob)
            image = Image.open(image_stream)

            # Perform OCR using pytesseract
            ocr_text = pytesseract.image_to_string(image)

            return ocr_text
        except Exception as e:
            self.logger.error(f"Error during image processing: {e}")
            return ""



import io
from typing import AsyncGenerator
from botocore.exceptions import (
    NoCredentialsError,
)
from querent.common.types.collected_bytes import (
    CollectedBytes,
)  # Import for handling authentication errors
from querent.config.collector.collector_config import CollectorBackend, S3CollectConfig
from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.common.uri import Uri
import boto3

from querent.logging.logger import setup_logger


class AWSCollector(Collector):
    def __init__(self, config: S3CollectConfig, prefix: str):
        self.bucket_name = config.bucket
        self.region = config.region
        self.access_key = config.access_key
        self.secret_key = config.secret_key
        self.chunk_size = 1024  # Default value
        if config.chunk and config.chunk.isdigit():
            self.chunk_size = int(config.chunk)
        self.logger = setup_logger(__name__, "AWSCollector")

    async def connect(self):
        # Initialize the S3 client with proper error handling for credentials
        try:
            self.s3_client = boto3.client(
                "s3",
                aws_access_key_id=self.access_key,
                aws_secret_access_key=self.secret_key,
                region_name=self.region,
            )
        except NoCredentialsError:
            raise Exception("AWS credentials are not set properly.")

    async def disconnect(self):
        # No asynchronous disconnect needed for boto3
        pass

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        if not self.s3_client:
            await self.connect()

        try:
            response = self.s3_client.list_objects_v2(Bucket=self.bucket_name)

            for obj in response.get("Contents", []):
                file = self.download_object_as_byte_stream(obj["Key"])
                async for chunk in self.read_chunks(file):
                    yield CollectedBytes(file=obj["Key"], data=chunk, error=None)
                yield CollectedBytes(file=obj["Key"], data=None, error=None, eof=True)

        except PermissionError as exc:
            self.logger.error(f"Getting Permission Error on file {file}, as {exc}")
        except OSError as exc:
            self.logger.error(f"Getting OS Error on file {file}, as {exc}")
        finally:
            await self.disconnect()  # Disconnect when done

    async def read_chunks(self, file):
        while True:
            chunk = file.read(self.chunk_size)
            if not chunk:
                break
            yield chunk

    def download_object_as_byte_stream(self, object_key):
        byte_stream = io.BytesIO()
        self.s3_client.download_fileobj(self.bucket_name, object_key, byte_stream)
        byte_stream.seek(0)  # Rewind the stream to the beginning
        return byte_stream


class AWSCollectorFactory(CollectorFactory):
    def backend(self) -> CollectorBackend:
        return CollectorBackend.S3

    def resolve(self, uri: Uri, config: S3CollectConfig) -> Collector:
        return AWSCollector(config, uri)



import json
from typing import AsyncGenerator

import aiofiles
from querent.common.types.collected_bytes import CollectedBytes
from querent.config.collector.collector_config import GcsCollectConfig
from querent.config.collector.collector_config import CollectorBackend
from querent.collectors.collector_base import Collector
from querent.collectors.collector_factory import CollectorFactory
from querent.common.uri import Uri
from querent.common import common_errors
from google.cloud import storage
from dotenv import load_dotenv

from querent.logging.logger import setup_logger

load_dotenv()


class GCSCollector(Collector):
    def __init__(self, config: GcsCollectConfig):
        self.bucket_name = config.bucket
        self.credentials = json.loads(config.credentials)
        self.client = None
        self.chunk_size = 1024
        if config.chunk and config.chunk.isdigit():
            self.chunk_size = int(config.chunk)
        self.logger = setup_logger(__name__, "GCSCollector")

    async def connect(self):
        if not self.client:
            try:
                self.client = storage.Client.from_service_account_info(self.credentials)
            except ConnectionError as exc:
                self.logger.error(f"Error connecting to GCS: {exc}")
                raise common_errors.ConnectionError(
                    "Please pass the credentials"
                ) from exc

    async def disconnect(self):
        try:
            if self.client is not None:
                self.client.close()
                self.client = None
        except Exception as exc:
            self.logger.error(f"Error disconnecting from GCS: {exc}")

    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        # Make sure to connect the client before using it
        if not self.client:
            await self.connect()

        try:
            bucket = self.client.get_bucket(self.bucket_name)
            blobs = list(bucket.list_blobs())  # Convert to a list
            for blob in blobs:
                async for chunk in self.stream_blob(blob):
                    yield CollectedBytes(file=blob.name, data=chunk, error=None)
                yield CollectedBytes(file=blob.name, data=None, error=None, eof=True)
        except Exception as e:
            # Handle exceptions gracefully, e.g., log the error
            self.logger.error(f"Error connecting to GCS: {e}")
        finally:
            # Disconnect the client when done
            await self.disconnect()

    async def stream_blob(self, blob):
        try:
            with blob.open("rb") as blob_file:
                while True:
                    chunk = blob_file.read(self.chunk_size)
                    if not chunk:
                        break
                    yield chunk
        except PermissionError as exc:
            raise common_errors.PermissionError(
                f"Unable to open this file {blob_file}, getting error as {exc}"
            ) from exc
        except OSError as exc:
            raise common_errors.OSError(
                f"Getting OS Error on file {blob_file}, as {exc}"
            ) from exc
        except Exception as exc:
            raise common_errors.UnknownError(
                f"Getting OS Error on file {blob_file}, as {exc}"
            ) from exc


class GCSCollectorFactory(CollectorFactory):
    def backend(self) -> CollectorBackend:
        return CollectorBackend.Gcs

    def resolve(self, uri: Uri, config: GcsCollectConfig) -> Collector:
        return GCSCollector(config)



from enum import Enum


class ConfigKey(Enum):
    QUERENT_ID = "querent_id"
    COLLECTORS = "collectors"
    ENGINES = "engines"
    RESOURCE = "resource"
    VERSION = "version"
    QUERENT_NAME = "querent_name"
    WORKFLOW = "workflow"
    # Add more keys as needed



import json
from typing import Any, List, Tuple
from querent.kg.rel_helperfunctions.embedding_store import EmbeddingStore
from querent.kg.rel_helperfunctions.questionanswer_llama2 import QASystem
from querent.kg.rel_helperfunctions.rag_retriever import RAGRetriever
from querent.kg.rel_helperfunctions.rel_normalize import TextNormalizer
from querent.logging.logger import setup_logger
from querent.config.core.opensource_llm_config import Opensource_LLM_Config
from llama_cpp import LlamaGrammar

import ast

"""
    A class for extracting relationships from triples and processing them for various representations.

    This class includes methods for validating triples, generating embeddings, processing tokens, 
    normalizing and building indices for triples, creating semantic triples, extracting relationships, 
    and trimming triples.

    Attributes:
        config (Opensource_LLM_Config): Configuration settings for the relationship extractor.
        logger (Logger): Logger for logging messages and errors.
        create_emb (EmbeddingStore): An instance of EmbeddingStore for generating embeddings.
        qa_system (QASystem): A question-answering system for extracting relationships.
        rag_approach (bool): A flag indicating whether to use the RAG approach for retrieval.
        rag_retriever (RAGRetriever): A RAG retriever for document retrieval, used if rag_approach is True.
        sub_tasks (list): List of dynamic sub-tasks for processing.

    Methods:
        validate(data) -> bool:
            Validates the input data to ensure it's in the correct format for processing.
        generate_embeddings(payload):
            Generates embeddings for the given payload containing triples.
        process_tokens(payload):
            Processes tokens in the given payload and extracts relationships.
        normalizetriples_buildindex(triples):
            Normalizes the given triples and builds an index for them.
        create_semantic_triple(input1, input2):
            Creates a semantic triple from the given inputs.
        extract_relationships(triples):
            Extracts relationships from the given triples.
        trim_triples(data):
            Trims the given data to a more concise format.
    """

class RelationExtractor():
    def __init__(self, config: Opensource_LLM_Config):  
        self.logger = setup_logger(config.logger, "RelationshipExtractor")
        try:
            super().__init__()
            self.config = config
            self.create_emb = EmbeddingStore(vector_store_path=config.vector_store_path)
            self.qa_system = QASystem(
                rel_model_path=config.model_path,
                rel_model_type=config.model_type,
                )
            self.grammar = LlamaGrammar.from_file(file = config.grammar_file_path)
            self.rag_approach = config.rag_approach
            if  self.rag_approach == True:
                self.rag_retriever = RAGRetriever(
                faiss_index_path=config.get_faiss_index_path(),
                emb_model_name=config.emb_model_name,
                embedding_store=self.create_emb,
                logger=self.logger)
            self.is_confined_search = config.is_confined_search
        except Exception as e:
            self.logger.error(f"Initialization failed: {e}")
            raise Exception(f"Initialization failed: {e}")
        
    def validate(self, data) -> bool:
        try:
            if not data:
                self.logger.error(
                    f"Invalid {self.__class__.__name__} configuration. Empty List Error"
                )
                return False

            if not isinstance(data, list):
                self.logger.error(
                    f"Invalid {self.__class__.__name__} configuration. Incorrect Format Error: Not a list"
                )
                return False

            item = data[0]
            if not isinstance(item, tuple) or len(item) != 3:
                self.logger.error(
                    f"Invalid {self.__class__.__name__} configuration. Incorrect Format Error: Item is not a triple"
                )
                return False

            if not (
                isinstance(item[0], str)
                and isinstance(item[2], str)
                and isinstance(item[1], str)
            ):
                self.logger.error(
                    f"Invalid {self.__class__.__name__} configuration. Incorrect Format Error: Incorrect item format"
                )
                return False

            return True
        except Exception as e:
            self.logger.error(f"Error in validation: {e}")
            return False
    
    def process_tokens(self, payload):
        try:
            triples = payload
            trimmed_triples = self.normalizetriples_buildindex(triples)
            if self.rag_approach == True:
                self.rag_retriever.build_faiss_index(trimmed_triples)
            relationships = self.extract_relationships(triples)
        
            return relationships
        
        except Exception as e:
            self.logger.error(f"Error in processing event: {e}")
            raise Exception(f"Invalid in processing event: {e}")

    def normalizetriples_buildindex(self, triples):
        try:
            if not self.validate(triples):
                self.logger.error("Invalid triples for relationship extractor")
                raise ValueError("Invalid triples for relationship extractor")
            normalizer = TextNormalizer()
            normalized_triples = normalizer.normalize_triples(triples)
            trimmed_triples = self.trim_triples(normalized_triples)
            
            return trimmed_triples
        
        except Exception as e:
            self.logger.error(f"Error in normalizing/building index: {e}")
            raise Exception(f"Error in normalizing/building index: {e}")
    
    def create_semantic_triple(self, input1, input2):
        try:   
            if not input1.get("predicate") or not input1.get("subject") or not input1.get("object"):
                raise ValueError("Missing 'subject', 'predicate', or 'object' in llm output") 
            input2_data = ast.literal_eval(input2)
            triple = (
                input1.get("subject",""),
                json.dumps({
                    "predicate": input1.get("predicate",""),
                    "predicate_type": input1.get("predicate_type","Unlabeled"),
                    "context": input2_data.get("context", ""),
                    "file_path": input2_data.get("file_path", ""),
                    "subject_type": input1.get("subject_type","Unlabeled"),
                    "object_type": input1.get("object_type","Unlabeled")
                }),
                input1.get("object","")
            )
            # else:
            #         if input1.get("subject","").lower() in input2_data.get("entity1_nn_chunk","").lower or input2_data.get("entity1_nn_chunk","").lower in input1.get("subject","").lower() or input2_data.get("entity1_nn_chunk","").lower == input1.get("subject","").lower():
            #             triple = (
            #             input1.get("subject",""),
            #             json.dumps({
            #                 "predicate": input1.get("predicate",""),
            #                 "predicate_type": input1.get("predicate_type","Unlabeled"),
            #                 "context": input2_data.get("context", ""),
            #                 "file_path": input2_data.get("file_path", ""),
            #                 "subject_type": input2_data.get("entity1_label","Unlabeled"),
            #                 "object_type": input2_data.get("entity2_label","Unlabeled")
            #             }),
            #             input1.get("object","")
            #         )
            #         else:
            #             triple = (
            #             input1.get("subject",""),
            #             json.dumps({
            #                 "predicate": input1.get("predicate",""),
            #                 "predicate_type": input1.get("predicate_type","Unlabeled"),
            #                 "context": input2_data.get("context", ""),
            #                 "file_path": input2_data.get("file_path", ""),
            #                 "subject_type": input2.get("entity2_label","Unlabeled"),
            #                 "object_type": input1.get("entity1_label","Unlabeled")
            #             }),
            #             input1.get("object","")
                    # )
            return triple
        except Exception as e:
            self.logger.error(f"Error in creating semantic triple: {e}")
            raise Exception(f"Error in creating semantic triple: {e}")
        
    def replace_entities(self, text, entity1, entity2):
        data = json.loads(text)
        if data.get("subject", "").strip() == "Entity 1":
            data["subject"] = entity1
        elif data.get("subject", "").strip() == "Entity 2":
            data["subject"] = entity2

        if data.get("object", "").strip() == "Entity 1":
            data["object"] = entity1
        elif data.get("object", "").strip() == "Entity 2":
            data["object"] = entity2

        return data

    def extract_relationships(self, triples):
        try:
            self.logger.info(f"Length of identified triples {len(triples)}")
            updated_triples = []
            for _, predicate_str, _ in triples:
                documents=[]
                data = json.loads(predicate_str)
                context = data['context']
                predicate = predicate_str if isinstance(predicate_str, dict) else json.loads(predicate_str)
                if self.rag_approach == True:
                    db = self.rag_retriever.load_faiss_index()
                    prompt=("What is the relationship between {entity1} and the Object is {entity2}.").format(entity1=predicate.get('entity1_nn_chunk', ''), entity2=predicate.get('entity2_nn_chunk', ''))
                    top_docs = self.rag_retriever.retrieve_documents(db, prompt=prompt)
                    documents = top_docs
                else:
                    if not self.config.qa_template:
                        query = """Please analyze the provided context and two entities. Use this information to answer the users query below.
Context: {context}
Entity 1: {entity1} and Entity 2: {entity2}
Query: In a semantic triple (Subject, Predicate & Object) framework, determine which of the above entity is the subject and which is the object based on the context along with the predicate between these entities. Please also identify the subject type, object type & predicate type.
Answer:""".format(context = context, entity1=predicate.get('entity1_nn_chunk', ''), entity2=predicate.get('entity2_nn_chunk', ''))   
                    else:
                        query = self.config.qa_template.format(context = context, entity1=predicate.get('entity1_nn_chunk', ''), entity2=predicate.get('entity2_nn_chunk', ''))    
                    answer_relation = self.qa_system.ask_question(prompt=query, llm=self.qa_system.llm, grammar=self.grammar)
                    try:
                        choices_text = answer_relation['choices'][0]['text']
                        answer_relation = self.replace_entities(choices_text,entity1=predicate.get('entity1_nn_chunk', ''), entity2=predicate.get('entity2_nn_chunk'))
                        updated_triple= self.create_semantic_triple(answer_relation, predicate_str)
                        updated_triples.append(updated_triple)
                    except Exception as e:
                        continue
            return updated_triples
        except Exception as e:
            self.logger.error(f"Error in extracting relationships: {e}")

    def trim_triples(self, data):
        try:
            trimmed_data = []
            for entity1, predicate, entity2 in data:
                predicate_dict = json.loads(predicate)
                trimmed_predicate = {
                    'context': predicate_dict.get('context', ''),
                    'entity1_nn_chunk': predicate_dict.get('entity1_nn_chunk', ''),
                    'entity2_nn_chunk': predicate_dict.get('entity2_nn_chunk', ''),
                    'file_path': predicate_dict.get('file_path', '')
                }
                trimmed_data.append((entity1, trimmed_predicate, entity2))

            return trimmed_data
        
        except Exception as e:
            self.logger.error(f"Error in trimming triples: {e}")
            raise Exception(f'Error in trimming triples: {e}')
  



from abc import ABC, abstractmethod
from typing import AsyncGenerator

from querent.common.types.collected_bytes import CollectedBytes


class Collector(ABC):
    @abstractmethod
    async def connect(self):
        raise NotImplementedError

    @abstractmethod
    async def poll(self) -> AsyncGenerator[CollectedBytes, None]:
        raise NotImplementedError

    @abstractmethod
    async def disconnect(self):
        raise NotImplementedError



"""File to start workflow"""
import asyncio
from querent.common.types.querent_queue import QuerentQueue
from querent.config.config import Config
from querent.config.workflow.workflow_config import WorkflowConfig
from querent.config.collector.collector_config import CollectorConfig
from querent.config.core.llm_config import LLM_Config
from querent.config.core.gpt_llm_config import GPTConfig
from querent.collectors.collector_resolver import CollectorResolver
from querent.common.uri import Uri
from querent.ingestors.ingestor_manager import IngestorFactoryManager
from querent.core.transformers.bert_ner_opensourcellm import BERTLLM
from querent.core.transformers.gpt_llm_gpt_ner import GPTNERLLM
from querent.common.types.querent_event import EventType
from querent.querent.querent import Querent
from querent.querent.resource_manager import ResourceManager
from querent.common.types.ingested_tokens import IngestedTokens

async def start_workflow(config_dict: dict):
    #Start the workflow
    workflow_config = config_dict.get("workflow")
    workflow = WorkflowConfig(config_source=workflow_config)
    collector_configs = config_dict.get("collectors", [])
    collectors = []
    for collector_config in collector_configs: 
        collectors.append(CollectorConfig(config_source=collector_config).resolve())

    engine_configs = config_dict.get("engines", [])
    engines = []
    for engine_config in engine_configs:
        engine_config_source = engine_config.get("config",{})
        if engine_config["name"] == "knowledge_graph_using_openai":
            engines.append(GPTConfig(config_source = engine_config_source))
        elif engine_config["name"] == "knowledge_graph_using_llama2_v1":
            engines.append(LLM_Config(config_source=engine_config_source))
    config_dict["engines"] = engines 
    config_dict["collectors"] = collectors 
    config_dict["workflow"] = workflow
    config = Config(config_source=config_dict)

    workflows = {"knowledge_graph_using_openai": start_gpt_workflow,
               "knowledge_graph_using_llama2_v1": start_llama_workflow}

    workflow = workflows.get(config.workflow.name)
    result_queue = QuerentQueue()
    state_queue = QuerentQueue()
    collector_tasks = asyncio.create_task(start_collectors(config))
    engine_tasks = asyncio.create_task(workflow(ResourceManager(), config, result_queue, state_queue))
    await asyncio.gather(collector_tasks, engine_tasks)
    print("Workflow is finished. All events have been released.")


async def start_collectors(config: Config):
    collectors = []
    for collector_config in config.collectors:
        uri = Uri(collector_config.uri)
        collectors.append(CollectorResolver().resolve(uri=uri, config = collector_config))

    for collector in collectors:
        await collector.connect()

    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=None,
        tokens_feader= config.workflow.tokens_feader
    )

    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())

    await ingest_task


async def start_gpt_workflow(resource_manager: ResourceManager, config: Config, result_queue: QuerentQueue,  state_queue: QuerentQueue):
    llm_instance = GPTNERLLM(result_queue, config.engines[0])

    llm_instance.subscribe(EventType.Graph, config.workflow.event_handler)
    querent = Querent(
        [llm_instance],
        resource_manager=resource_manager,
    )
    querent_task = asyncio.create_task(querent.start())
    token_feeder = asyncio.create_task(receive_token_feeder(resource_manager=resource_manager, config=config, result_queue=result_queue, state_queue=llm_instance.state_queue))
    await asyncio.gather(querent_task, token_feeder)


#Config workflow channel for setting termination event
async def receive_token_feeder(resource_manager: ResourceManager, config: Config, result_queue: QuerentQueue, state_queue: QuerentQueue):
    await asyncio.sleep(180)
    while not resource_manager.querent_termination_event.is_set():
        tokens = config.workflow.tokens_feader.receive_tokens_in_python()
        message_state = config.workflow.channel.receive_in_python()
        if tokens is not None:
            
            ingested_tokens = IngestedTokens(file=tokens.get("file", None),
                                    data=tokens.get("data", None))
            #we will get a dictionary here
            await result_queue.put(ingested_tokens)

        else:
            break
        if message_state is not None:
            state_queue.put_nowait(message_state)
    
    await result_queue.put(None)

async def start_llama_workflow(resource_manager: ResourceManager, config: Config, result_queue: QuerentQueue,  state_queue: QuerentQueue):

    llm_instance = BERTLLM(result_queue, config.engines[0])

    llm_instance.subscribe(EventType.Graph, config.workflow.event_handler)
    querent = Querent(
        [llm_instance],
        resource_manager=resource_manager,
    )
    querent_task = asyncio.create_task(querent.start())
    token_feeder = asyncio.create_task(receive_token_feeder(resource_manager=resource_manager, config=config, result_queue=result_queue, state_queue=llm_instance.state_queue))
    await asyncio.gather(querent_task, token_feeder)

async def start_ingestion(config_dict: dict):
    collectors = []
    collector_configs = config_dict.get("collectors", [])
    for collector_config in collector_configs: 
        collectors.append(CollectorConfig(config_source=collector_config).resolve())
    workflow_config = config_dict.get("workflow")
    workflow = WorkflowConfig(config_source=workflow_config)
    engine_configs = config_dict.get("engines", [])
    engines = []
    for engine_config in engine_configs:
        engine_config_source = engine_config.get("config",{})
        if engine_config["name"] == "knowledge_graph_using_openai":
            engines.append(GPTConfig(config_source = engine_config_source))
        elif engine_config["name"] == "knowledge_graph_using_llama2_v1":
            engines.append(LLM_Config(config_source=engine_config_source))
    config_dict["engines"] = engines 
    config_dict["collectors"] = collectors 
    config_dict["workflow"] = workflow
    config = Config(config_source=config_dict)
    collectors = []
    for collector_config in config.collectors:
        uri = Uri(collector_config.uri)
        collectors.append(CollectorResolver().resolve(uri=uri, config = collector_config))

    for collector in collectors:
        await collector.connect()

    ingestor_factory_manager = IngestorFactoryManager(
        collectors=collectors, result_queue=None,
        tokens_feader= config.workflow.tokens_feader
    )

    ingest_task = asyncio.create_task(ingestor_factory_manager.ingest_all_async())

    await ingest_task
    

async def start_workflow_engine(config_dict: Config):
    result_queue = QuerentQueue()
    state_queue = QuerentQueue()
    workflows = {"knowledge_graph_using_openai": start_gpt_workflow,
               "knowledge_graph_using_llama2_v1": start_llama_workflow}
    config = Config(config_dict)
    workflow = workflows.get(config.workflow.name)
    resource_manager = ResourceManager()
    receiving_tokens = asyncio.create_task(receive_token_feeder(resource_manager, config, result_queue, state_queue))
    starting_engine = asyncio.create_task(workflow(resource_manager, config, result_queue, state_queue))
    await asyncio.gather(receiving_tokens, starting_engine)


from typing import List, AsyncGenerator
import uuid
from PIL import Image
import pybase64
import pypdf
from querent.common import common_errors
from querent.common.types.collected_bytes import CollectedBytes
from querent.common.types.ingested_images import IngestedImages
from querent.ingestors.base_ingestor import BaseIngestor
from querent.ingestors.email.email_reader import EmailReader
from querent.ingestors.ingestor_factory import IngestorFactory
from querent.logging.logger import setup_logger
from querent.processors.async_processor import AsyncProcessor
from querent.config.ingestor.ingestor_config import IngestorBackend
from querent.common.types.ingested_tokens import IngestedTokens
import email
import pytesseract
from io import BytesIO
import io


class EmailIngestorFactory(IngestorFactory):
    SUPPORTED_EXTENSIONS = {"email", "eml"}

    async def supports(self, file_extension: str) -> bool:
        return file_extension.lower() in self.SUPPORTED_EXTENSIONS

    async def create(
        self, file_extension: str, processors: List[AsyncProcessor]
    ) -> BaseIngestor:
        if not await self.supports(file_extension):
            return None
        return EmailIngestor(processors)


class EmailIngestor(BaseIngestor):
    def __init__(self, processors: List[AsyncProcessor]):
        super().__init__(IngestorBackend.Email)
        self.processors = processors
        self.logger = setup_logger(__name__, "EmailIngestor")
        self.email_reader = EmailReader()

    async def ingest(
        self, poll_function: AsyncGenerator[CollectedBytes, None]
    ) -> AsyncGenerator[IngestedTokens, None]:
        collected_bytes = b""
        current_file = None
        try:
            async for chunk_bytes in poll_function:
                if chunk_bytes.is_error() or chunk_bytes.is_eof():
                    continue

                if current_file is None:
                    current_file = chunk_bytes.file
                elif current_file != chunk_bytes.file:
                    email = await self.extract_and_process_email(
                        CollectedBytes(file=current_file, data=collected_bytes)
                    )
                    yield IngestedTokens(
                        file=current_file,
                        data=email,  # Wrap line in a list
                        error=None,
                    )
                    yield IngestedTokens(
                        file=current_file,
                        data=None,
                        error=None,
                    )
                    collected_bytes = b""
                    current_file = chunk_bytes.file
                collected_bytes += chunk_bytes.data
        except Exception as e:
            yield IngestedTokens(file=current_file, data=None, error=f"Exception: {e}")
        finally:
            if current_file is not None:
                email = await self.extract_and_process_email(
                    CollectedBytes(file=current_file, data=collected_bytes)
                )
                yield IngestedTokens(
                    file=current_file,
                    data=email,  # Wrap line in a list
                    error=None,
                )
                yield IngestedTokens(
                    file=current_file,
                    data=None,
                    error=None,
                )

    async def extract_and_process_email(
        self, collected_bytes: CollectedBytes
    ) -> List[str]:
        text = await self.extract_text_from_email(collected_bytes)
        processed_text = await self.process_data(text)
        return processed_text

    async def extract_text_from_email(self, collected_bytes: CollectedBytes) -> str:
        text = ""
        try:
            msg = email.message_from_bytes(collected_bytes.data)
            email_msg = {}
            (
                email_msg["From"],
                email_msg["To"],
                email_msg["Date"],
                email_msg["Subject"],
            ) = self.email_reader.obtain_header(msg)
            if msg.is_multipart():
                for part in msg.walk():
                    content_type = part.get_content_type()
                    try:
                        body = part.get_payload(decode=True)
                    except Exception as e:
                        continue
                    if body is None:
                        continue
                    text += await self.handle_sub_part(part)
                    text += "\n"
            else:
                content_type = msg.get_content_type()
                body = msg.get_payload(decode=True).decode()
                if content_type == "text/plain":
                    text = self.email_reader.clean_email_body(body)
        except Exception as e:
            self.logger.error(f"Error extracting text from email: {e}")
        return text

    async def handle_sub_part(self, sub_part):
        sub_content_type = sub_part.get_content_type()
        sub_content_disposition = str(sub_part.get("Content-Disposition"))
        if (
            sub_content_type == "text/plain"
            and "attachment" not in sub_content_disposition
        ):
            return self.email_reader.clean_email_body(sub_part.get_payload())
        elif "attachment" in sub_content_disposition:
            # Handle attachment as needed
            return await self.handle_attachment(sub_part)
        elif sub_content_type == "text/html":
            # Handle HTML content as needed
            # You can choose to ignore or process it
            return ""
        else:
            # Handle other content types as needed
            return ""

    async def handle_attachment(self, attachment_part) -> str:
        # Get attachment data and type
        attachment_data = attachment_part.get_payload(decode=True)
        attachment_type = attachment_part.get_content_type()

        # Check attachment type and handle accordingly
        if attachment_type.startswith("image/"):
            return await self.handle_image_attachment(attachment_data)
        elif attachment_type == "application/pdf":
            return await self.handle_pdf_attachment(attachment_data)
        else:
            # Handle other attachment types as needed
            self.logger.warning(f"Unsupported attachment type: {attachment_type}")

    async def handle_image_attachment(self, image_data: bytes) -> str:
        try:
            ocr_text = await self.get_ocr_from_image(image_data)
            return ocr_text
        except Exception as e:
            self.logger.error(f"Error handling image attachment: {e}")

    async def handle_pdf_attachment(self, pdf_data) -> str:
        try:
            pdf_text = await self.extract_and_process_pdf(pdf_data)
        except Exception as e:
            self.logger.error(f"Error handling PDF attachment: {e}")
        return pdf_text

    async def get_ocr_from_image(self, image):
        """Implement this to return ocr text of the image"""
        image = Image.open(io.BytesIO(image))
        text = pytesseract.image_to_string(image)
        return str(text).encode("utf-8").decode("unicode_escape")

    async def extract_and_process_pdf(self, pdf_data: bytes) -> str:
        pdf_text = ""
        try:
            path = BytesIO(pdf_data)
            loader = pypdf.PdfReader(path)

            for _, page in enumerate(loader.pages):
                text = page.extract_text()
                pdf_text += text + "\n"
                pdf_text += await self.extract_images_and_ocr(page)

        except TypeError as exc:
            self.logger.error(f"Exception while extracting email {exc}")
        except Exception as exc:
            self.logger.error(f"Exception while extracting email {exc}")
        return pdf_text

    async def extract_images_and_ocr(self, page) -> str:
        ocr_text = ""
        try:
            for image_path in page.images:
                ocr_text += await self.get_ocr_from_image(image_path)
        except Exception as e:
            self.logger.error(f"Error extracting images and OCR: {e}")
        return ocr_text

    async def process_data(self, text: str) -> str:
        if self.processors is None or len(self.processors) == 0:
            return text
        try:
            processed_data = text
            for processor in self.processors:
                processed_data = await processor.process_text(processed_data)
            return processed_data
        except Exception as e:
            self.logger.error(f"Error while processing text: {e}")



from querent.graph.errors import InvalidParameter
from querent.graph.utils import URI, BNode, Literal


class Subject:
    def __init__(self, s):
        if not isinstance(s, (URI, BNode)):
            raise InvalidParameter("Subject needs to be URI or BNode")
        self._resource = dict()
        self._s = s
        self._size_len = 0

    def add_property(self, p, o, reify=None):
        """

        :param p: URI, predicate
        :param o: [URI, BNode, Literal, Subject], object
        :param reify: Optional[Reification, Tuple[URI, Optional[URI or BNode],
                        Tuple[URI, URI, Optional[URI, BNode]]
                    None -> don't reify
                    Reification -> reify with Reification specification
                    URI -> reify with specific predicate
                    (URI, Optional[Term]) -> reify, and use provided statement term or BNode
                    (URI, URI, Optional[Term]) -> reify with two predicates, statement term or BNode
        :return: statement term if reify else None
        """
        if not isinstance(p, URI):
            raise InvalidParameter("Predict needs to be URI")
        if not self.__is_valid_object(o):
            raise InvalidParameter(
                "Object needs to be URI or BNode or Literal or Triple"
            )

        if p not in self._resource:
            self._resource[p] = set([])
        self._resource[p].add(o)

        if reify:
            self.add_property(reify.p1, reify)
            if isinstance(reify, Reification):
                reify.add_property(reify.p2, o)
            return reify

    def remove_property(self, p, o=None):
        if not isinstance(p, URI):
            raise InvalidParameter("Predicate needs to be URI")
        if o and not self.__is_valid_object(o):
            raise InvalidParameter(
                "Object needs to be URI or BNode or Literal or Triple"
            )

        try:
            if not o:
                del self._resource[p]
            else:
                self._resource[p].remove(o)
                if not self._resource[p]:
                    del self._resource[p]
        except KeyError:
            return False

        return True

    @property
    def subject(self):
        return self._s

    @staticmethod
    def __is_valid_object(o):
        return isinstance(o, (URI, BNode, Literal, Subject))

    def __iter__(self):
        for p, os in self._resource.items():
            for o in os:
                yield self._s, p, o

    def __next__(self):
        return self.__iter__()

    def _calculate_memory_usage(self):
        return sum(
            [
                len(str(self._s)),
                sum(
                    [
                        len(str(p)) + len(str(o))
                        for p, os in self._resource.items()
                        for o in os
                    ]
                ),
            ]
        )


class Reification(Subject):
    def __init__(self, p1, p2=None, statement=None):
        """
        Make reification.

        :param p1: URI, predicate for reification
        :param p2: Optional[URI],
                    URI -> second predicate for reification
                    None -> reuse p1 as second predicate
        :param statement: Optional[Union[URI, BNode]],
                    URI or BNode -> use provided node as statement term
                    None -> generate a new BNode as statement term
        """
        if isinstance(p1, URI):
            self.p1 = p1
        else:
            raise InvalidParameter("Reification predicate needs to be a valid URI")

        if p2 is None:
            self.p2 = p1
        elif isinstance(p2, URI):
            self.p2 = p2
        else:
            raise InvalidParameter(
                "Reification second predicate needs to be a valid URI or None"
            )

        if statement is None:
            self.statement = BNode()
        elif isinstance(statement, (URI, BNode)):
            self.statement = statement
        else:
            raise InvalidParameter(
                "Reification statement term needs to be a valid URI or BNode or None"
            )
        super().__init__(self.statement)



from rdflib import OWL, RDF
from querent.config.graph_config import GraphConfig

from querent.graph.graph import QuerentGraph


class GraphOntology(QuerentGraph):
    def __init__(self, graph_config: GraphConfig):
        super().__init__(graph_config)

    def is_valid(self, s_types, p, o_types):
        """
        Check if the triple is valid according to the schema
        """
        return True

    def add_subject_ontologies(self, subjects):
        """
        Add a list of subjects to the graph
        :param subjects: List[Subject]
        :return:
        """
        self.add_subjects(subjects)

    @property
    def object_properties(self):
        """
        Return all the defined ObjectProperty
        :return: Set[URIRef]
        """
        properties = []
        for property_ in self.graph.subjects(RDF.type, OWL.ObjectProperty):
            properties.append(property_)
        return set(properties)

    @property
    def datatype_properties(self):
        """
        Return all the defined DatatypeProperty
        :return: Set[URIRef]
        """
        properties = []
        for property_ in self.graph.subjects(RDF.type, OWL.DatatypeProperty):
            properties.append(property_)
        return set(properties)
